#include <string.h>
#include <gtk/gtk.h>

#include "gtkmdi.h"
#include "taskbar.h"

struct _GtkMdiTaskBar
{
	GtkToolbar parent_instance;
	GtkMdi *mdi;
	gboolean ignore_toggled;
};

static void gtk_mdi_taskbar_class_init (GtkMdiTaskBarClass *);
static void gtk_mdi_taskbar_init (GtkMdiTaskBar *);
static void gtk_mdi_taskbar_finalize (GObject *);
static void gtk_mdi_taskbar_add (GtkContainer *, GtkWidget *);
static void gtk_mdi_taskbar_remove (GtkContainer *, GtkWidget *);

static GtkWidget * gtk_mdi_taskbar_find_window (GtkMdiTaskBar *,
						GtkMdiWindow *);
static GtkWidget * gtk_mdi_taskbar_find_window_by_id (GtkMdiTaskBar *,
						GtkMdiWindowId);

static void _gtk_mdi_taskbar_window_title_changed (GtkMdi *, GtkMdiWindow *,
						gpointer);
static void _gtk_mdi_taskbar_window_icon_changed (GtkMdi *, GtkMdiWindow *, gpointer);
static void _gtk_mdi_taskbar_window_property_changed (GtkMdi *, GtkMdiWindow *,
					gchar *, gboolean, gboolean, gpointer);
static void _gtk_mdi_taskbar_window_set_deiconified (GtkMdi *, GtkMdiWindow *,
						gpointer);
static void _gtk_mdi_taskbar_window_set_iconified (GtkMdi *, GtkMdiWindow *,
						gpointer);
static void _gtk_mdi_taskbar_remove_window (GtkMdi *, GtkMdiWindowId, gpointer);
static void _gtk_mdi_taskbar_add_window (GtkMdi *, GtkMdiWindow *, gpointer);
static void _gtk_mdi_taskbar_window_has_get_focus (GtkMdi *, GtkMdiWindow *,
						gpointer);
static void _gtk_mdi_tb_change_window_state (GtkWidget *, gpointer);
static void _gtk_mdi_taskbar_window_remove_focus (GtkWidget *widget,
						gpointer data);

static GtkToolbar *parent_class;

GType
gtk_mdi_taskbar_get_type (void)
{
	static GType gtk_mdi_taskbar_type = 0;

	if (!gtk_mdi_taskbar_type)
	{
		static const GTypeInfo gtk_mdi_taskbar_info = {
			sizeof (GtkMdiTaskBarClass),
			NULL,
			NULL,
			(GClassInitFunc) gtk_mdi_taskbar_class_init,
			NULL,
			NULL,
			sizeof (GtkMdiTaskBar),
			0,
			(GInstanceInitFunc) gtk_mdi_taskbar_init
		};

		gtk_mdi_taskbar_type = g_type_register_static (
						GTK_TYPE_TOOLBAR,
						"GtkMdiTaskBar",
						&gtk_mdi_taskbar_info, 0);
	}

	return gtk_mdi_taskbar_type;
}

static void
gtk_mdi_taskbar_class_init (GtkMdiTaskBarClass *class)
{
	GObjectClass *o_class = (GObjectClass *)class;
	GtkContainerClass *c_class = (GtkContainerClass *)class;

	parent_class = g_type_class_peek_parent (class);

	//o_class->finalize = gtk_mdi_taskbar_finalize;
	c_class->add = gtk_mdi_taskbar_add;
	c_class->remove = gtk_mdi_taskbar_remove;
}

static void
gtk_mdi_taskbar_init (GtkMdiTaskBar *bar)
{
	bar->ignore_toggled = FALSE;
}

GtkWidget *
gtk_mdi_taskbar_new (GtkMdi *mdi)
{
	GtkWidget *new_tb;
	GtkMdiTaskBar *bar;

	g_return_val_if_fail (GTK_IS_MDI (mdi), NULL);

	new_tb = gtk_widget_new (GTK_MDI_TASKBAR_TYPE, NULL);

	bar = GTK_MDI_TASKBAR (new_tb);
	bar->mdi = mdi;

	g_signal_connect (bar->mdi, "window_opened",
			(GCallback)_gtk_mdi_taskbar_add_window, bar);
	g_signal_connect (bar->mdi, "window_closed",
			(GCallback)_gtk_mdi_taskbar_remove_window, bar);
	g_signal_connect (bar->mdi, "window_iconified",
			(GCallback)_gtk_mdi_taskbar_window_set_iconified,
			bar);
	g_signal_connect (bar->mdi, "window_deiconified",
			(GCallback)_gtk_mdi_taskbar_window_set_deiconified,
			bar);
	g_signal_connect (bar->mdi, "window_property_changed",
			(GCallback)_gtk_mdi_taskbar_window_property_changed,
			bar);
	g_signal_connect (bar->mdi, "window_icon_changed",
			(GCallback)_gtk_mdi_taskbar_window_icon_changed, bar);
	g_signal_connect (bar->mdi, "window_title_changed",
			(GCallback)_gtk_mdi_taskbar_window_title_changed, bar);
	g_signal_connect (bar->mdi, "window_focus_get",
			(GCallback)_gtk_mdi_taskbar_window_has_get_focus,
			bar);

	gtk_toolbar_set_style (GTK_TOOLBAR (bar), GTK_TOOLBAR_BOTH);
	gtk_toolbar_set_show_arrow (GTK_TOOLBAR (bar), TRUE);

	return new_tb;
}

static void
gtk_mdi_taskbar_add (GtkContainer *container, GtkWidget *widget)
{
	GtkToolItem *element;
	GtkMdiWindow *window;
	GtkMdiWindowId *id = g_new0 (GtkMdiWindowId, 1);

	g_return_if_fail (GTK_IS_MDI_TASKBAR (container));
	g_return_if_fail (GTK_IS_MDI_WINDOW (widget));

	window = GTK_MDI_WINDOW (widget);

	element = gtk_toggle_tool_button_new ();
	gtk_tool_button_set_icon_widget (GTK_TOOL_BUTTON (element),
			gtk_mdi_window_get_icon (GTK_MDI_WINDOW (widget)));

	gtk_tool_button_set_label_widget (GTK_TOOL_BUTTON (element),
		gtk_label_new (gtk_mdi_window_get_title (
			GTK_MDI_WINDOW (widget))));

	gtk_tool_item_set_tooltip (element, GTK_TOOLBAR (container)->tooltips,
			gtk_mdi_window_get_title (GTK_MDI_WINDOW (widget)),
			"");

	g_signal_connect (element, "toggled",
			(GtkSignalFunc)_gtk_mdi_tb_change_window_state,
			window);

	(*id) = window->id;
	g_object_set_data (G_OBJECT (element), "gtk_mdi_window", window);
	g_object_set_data (G_OBJECT (element), "gtk_mdi_window_id", id);

	if (GTK_CONTAINER_CLASS (parent_class)->add)
		GTK_CONTAINER_CLASS (parent_class)->add (container,
							GTK_WIDGET (element));
}

static void
gtk_mdi_taskbar_remove (GtkContainer *container, GtkWidget *widget)
{
	GtkWidget *element;

	g_return_if_fail (GTK_IS_MDI_TASKBAR (container));
	g_return_if_fail (GTK_IS_MDI_WINDOW (widget));

	element = gtk_mdi_taskbar_find_window (GTK_MDI_TASKBAR (container),
						GTK_MDI_WINDOW (widget));

	if (GTK_IS_TOGGLE_BUTTON (element))
	{
		gtk_container_remove (container, element);
	}
}

void
gtk_mdi_taskbar_append_window (GtkMdiTaskBar *bar, GtkMdiWindow *window)
{
	gtk_mdi_taskbar_add (GTK_CONTAINER (bar), GTK_WIDGET (window));
}

void
gtk_mdi_taskbar_remove_window (GtkMdiTaskBar *bar, GtkMdiWindowId id)
{
	GtkWidget *element;

	element = gtk_mdi_taskbar_find_window_by_id (bar, id);

	if (GTK_IS_TOGGLE_BUTTON (element))
	{
		gtk_container_remove (GTK_CONTAINER (bar), element);
	}
}

static GtkWidget *
gtk_mdi_taskbar_find_window (GtkMdiTaskBar *bar, GtkMdiWindow *window)
{
	GtkWidget *ret = NULL;
	GList *child;

	g_return_val_if_fail (GTK_IS_MDI_TASKBAR (bar), NULL);
	g_return_val_if_fail (GTK_IS_MDI_WINDOW (window), NULL);

	child = gtk_container_get_children (GTK_CONTAINER (bar));

	for ( ; child && !ret ; child = g_list_next (child))
	{
		GtkMdiWindow *win;
		GtkToggleToolButton *butt =
				GTK_TOGGLE_TOOL_BUTTON (child->data);

		win = g_object_get_data (G_OBJECT (butt), "gtk_mdi_window");
		if (win == window)
		{
			ret = GTK_WIDGET (butt);
		}
	}

	return ret;
}

static GtkWidget *
gtk_mdi_taskbar_find_window_by_id (GtkMdiTaskBar *bar, GtkMdiWindowId id)
{
	GtkWidget *ret = NULL;
	GList *child;

	g_return_val_if_fail (GTK_IS_MDI_TASKBAR (bar), NULL);

	child = gtk_container_get_children (GTK_CONTAINER (bar));

	for ( ; child && !ret ; child = g_list_next (child))
	{
		GtkMdiWindowId *tmp;
		GtkToggleToolButton *butt =
				GTK_TOGGLE_TOOL_BUTTON (child->data);

		tmp = g_object_get_data (G_OBJECT (butt), "gtk_mdi_window_id");
		if (tmp && (*tmp == id))
		{
			ret = GTK_WIDGET (butt);
		}
	}

	return ret;
}

void
_gtk_mdi_taskbar_add_window (GtkMdi *mdi, GtkMdiWindow *window, gpointer data)
{
	GtkMdiTaskBar *bar = (GtkMdiTaskBar *)data;

	g_return_if_fail (GTK_IS_MDI_TASKBAR (bar));
	g_return_if_fail (GTK_IS_MDI_WINDOW (window));

	gtk_mdi_taskbar_append_window (bar, window);
}

void
_gtk_mdi_taskbar_remove_window (GtkMdi *mdi, GtkMdiWindowId id, gpointer data)
{
	GtkMdiTaskBar *bar = (GtkMdiTaskBar *)data;

	g_return_if_fail (GTK_IS_MDI_TASKBAR (bar));

	gtk_mdi_taskbar_remove_window (bar, id);
}

void
_gtk_mdi_taskbar_window_set_iconified (GtkMdi *mdi, GtkMdiWindow *window,
					gpointer data)
{
	GtkMdiTaskBar *bar = (GtkMdiTaskBar *)data;
	GtkWidget *butt;

	g_return_if_fail (GTK_IS_MDI_TASKBAR (bar));
	g_return_if_fail (GTK_IS_MDI_WINDOW (window));

	butt = gtk_mdi_taskbar_find_window (bar, window);
	if (butt)
	{
		bar->ignore_toggled = TRUE;
		gtk_toggle_tool_button_set_active (
				GTK_TOGGLE_TOOL_BUTTON (butt), FALSE);
		bar->ignore_toggled = FALSE;
	}
}

void
_gtk_mdi_taskbar_window_has_get_focus (GtkMdi *mdi, GtkMdiWindow *window,
					gpointer data)
{
	GtkMdiTaskBar *bar = (GtkMdiTaskBar *)data;
	GtkWidget *butt;

	g_return_if_fail (GTK_IS_MDI_TASKBAR (bar));
	g_return_if_fail (GTK_IS_MDI_WINDOW (window));

	gtk_container_foreach (GTK_CONTAINER (bar),
				_gtk_mdi_taskbar_window_remove_focus, NULL);

	butt = gtk_mdi_taskbar_find_window (bar, window);
	if (butt)
	{
		gtk_mdi_window_set_has_focus (window, TRUE);
		bar->ignore_toggled = TRUE;
		gtk_toggle_tool_button_set_active (
				GTK_TOGGLE_TOOL_BUTTON (butt), TRUE);
		bar->ignore_toggled = FALSE;
	}
}

static void
_gtk_mdi_taskbar_window_remove_focus (GtkWidget *widget, gpointer data)
{
	GtkMdiWindow *window;

	g_return_if_fail (GTK_IS_TOGGLE_TOOL_BUTTON (widget));

	window = GTK_MDI_WINDOW (g_object_get_data (widget, "gtk_mdi_window"));
	gtk_mdi_window_set_has_focus (window, FALSE);
}

static void
_gtk_mdi_taskbar_window_set_deiconified (GtkMdi *mdi, GtkMdiWindow *window,
					gpointer data)
{
}

static void
_gtk_mdi_taskbar_window_property_changed (GtkMdi *mdi, GtkMdiWindow *window,
					gchar * name,
					gboolean old_value,
					gboolean value,
					gpointer data)
{
	GtkMdiTaskBar *bar = (GtkMdiTaskBar *)data;

	g_return_if_fail (GTK_IS_MDI_TASKBAR (bar));
	g_return_if_fail (GTK_IS_MDI_WINDOW (window));
	g_return_if_fail (name != NULL);

	if ((strcmp (name, "skip_taskbar") == 0) && value)
		gtk_mdi_taskbar_remove_window (bar, window->id);
	else if ((strcmp (name, "skip_taskbar") == 0) && !value)
		gtk_mdi_taskbar_append_window (bar, window);
	else if ((strcmp (name, "has_focus") == 0) && !value)
	{
		GtkWidget *element;

		element = gtk_mdi_taskbar_find_window (bar, window);
		if (element)
		{
			bar->ignore_toggled = TRUE;
			gtk_toggle_tool_button_set_active (element, FALSE);
			bar->ignore_toggled = FALSE;
		}
	}
}

static void
_gtk_mdi_taskbar_window_icon_changed (GtkMdi *mdi, GtkMdiWindow *window,
					gpointer data)
{
	GtkMdiTaskBar *bar = (GtkMdiTaskBar *)data;
	GtkWidget *butt;

	g_return_if_fail (GTK_IS_MDI_TASKBAR (bar));
	g_return_if_fail (GTK_IS_MDI_WINDOW (window));

	butt = gtk_mdi_taskbar_find_window (bar, window);
	if (butt)
	{
		gtk_container_remove (GTK_CONTAINER (butt), 
			gtk_tool_button_get_icon_widget (
					GTK_TOOL_BUTTON (butt)));
		gtk_tool_button_set_icon_widget (GTK_TOOL_BUTTON (butt), 
				gtk_mdi_window_get_icon (window));
	}
}

static void
_gtk_mdi_taskbar_window_title_changed (GtkMdi *mdi, GtkMdiWindow *window,
					gpointer data)
{
	GtkMdiTaskBar *bar = (GtkMdiTaskBar *)data;
	GtkWidget *butt;
	GtkWidget *label;

	g_return_if_fail (GTK_IS_MDI_TASKBAR (bar));
	g_return_if_fail (GTK_IS_MDI_WINDOW (window));

	butt = gtk_mdi_taskbar_find_window (bar, window);
	if (butt)
	{
		label = gtk_tool_button_get_label_widget (
						GTK_TOOL_BUTTON (butt));
		gtk_label_set_text (GTK_LABEL (label),
				gtk_mdi_window_get_title (window));
	}
}

static void
_gtk_mdi_tb_change_window_state (GtkWidget *button, gpointer data)
{
	GtkMdiWindow *window = GTK_MDI_WINDOW (data);
	GtkMdiTaskBar *bar;

	g_return_if_fail (GTK_IS_TOGGLE_TOOL_BUTTON (button));
	g_return_if_fail (GTK_IS_MDI_WINDOW (window));

	bar = gtk_widget_get_parent (button);

	if (GTK_IS_MDI_TASKBAR (bar) && bar->ignore_toggled) return;

	if (gtk_toggle_tool_button_get_active(GTK_TOGGLE_TOOL_BUTTON (button))
	&& (gtk_mdi_window_get_state (window) == GTK_MDI_WINDOW_ICONIFIED))
	{
		gtk_mdi_window_deiconify (window);
	}
	else if (!gtk_toggle_tool_button_get_active(GTK_TOGGLE_TOOL_BUTTON (button)) && gtk_mdi_window_has_focus (window))
		gtk_mdi_window_iconify (window);
	else if (gtk_toggle_tool_button_get_active(GTK_TOGGLE_TOOL_BUTTON (button)) && !gtk_mdi_window_has_focus (window))
		gtk_mdi_window_grab_focus (window);
}

